#pragma	config	OSC = HSPLL
#pragma config	WDT = ON, WDTPS = 256
/*
 * This program controls and monitors solar power battery arrays
 * 
 *  
 * USART2 		is the host comm port
 * USART1		aux comm port
 * Timer0		1 second clock
 * Timer1		background I/O clock
 * PORTA		analog inputs
 * adc0	vbatol				PIC Controller supply voltage to 5VDC regulator
 * adc1	solar				Charging voltage at battery from CC
 * adc2	currentin			50A AMPLOC sensor input from PV array
 * adc3	rawp1/inputpower	Voltage from PV array
 * adc4	rawp2/primarypower	Voltage a primary inverter battery
 * PORTF		analog inputs
 * adc5	current				300A AMPLOC sensor battery output to inverter
 * PORTD		switch input
 * PORTE		battery/charge relays
 * PORTJ		load/input relays
 * PORTB		external control i/o
 * PORTH0		run flasher led onboard.
 * PORTF		C40 charge controller led input
 * 2x16 LCD status panel and led status lights.
 * 
 * nsaspook@nsaspook.com Copyright 2010
 * Fairview, Oregon
 */

 /*
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * 
 * This application is designed for use with the
 * ET-BASE PIC8722 board, LCD display and XANTREX C40 charge controller.
 *  
 * RS-232 host commands 9600baud 8n1
 * A		send controller/battery status
 * B		select battery for modify commands, used with S M L
 * C		(NOT DONE YET)
 * D V M L	Calibrate ADC voltages, example DVMZ would adjust adc0 up 1 tick, DDDVLLZ would adjust adc2 down two ticks
 *			THIS FUNCTION USES THE EEPROM WRITE DURING INTERRUPT (TOO MANY FAST KEYPRESSES WITH CAUSE A LOCKUP)
 * F		Force current (none critical) charge routine to exit
 * K		lockup controller causing WDT timer to reboot
 * S M L	modify selected battery type, example: ZBBMZ would set battery 2 to M type.
 * Z		reset command parser
 * ?		controller attention needed/sent to host to request action (usually a status)
 * $		fuse blown error. (NOT DONE YET)
 * 
 */

#include <p18F8722.h>
#include <usart.h>
#include <xlcd.h>
#include <delays.h>
#include <string.h>
#include <stdlib.h>
#include <eep.h>
#include <stdio.h>
#include <adc.h>
#include <timers.h>

void rx_handler (void);

#define	TIMEROFFSET	26473				// timer0 16bit counter value for 1 second to overflow
#define BUF_SIZE 	128
#define FALSE		0
#define TRUE		1
#define	ON			0					// control relay states
#define OFF			1
#define BCRELAYS	PORTE
#define IORELAYS	PORTJ
#define	DIPSW		PORTD
#define	EXTIO		PORTB				// low nibble output, high nibble input
#define	CCLED		PORTCbits.RC0
#define	STFU		PORTCbits.RC1		// kill alarm
#define SOLAROFF	PORTJbits.RJ7
#define	CHARGER		PORTJbits.RJ5
#define	CHARGERLED	PORTJbits.RJ5
#define VENTFAN		PORTJbits.RJ6
#define ALARMOUT	PORTJbits.RJ4
#define BAT1		PORTEbits.RE0
#define BAT2		PORTEbits.RE1
#define BAT3		PORTEbits.RE2
#define BAT4		PORTEbits.RE3
#define	CHRG1		PORTEbits.RE4
#define	CHRG2		PORTEbits.RE5
#define	CHRG3		PORTEbits.RE6
#define	CHRG4		PORTEbits.RE7
#define	CCBYPASS	PORTJbits.RJ0
#define	BATLOAD		PORTJbits.RJ1
#define	CHRGVOLTS	PORTJbits.RJ2
#define	CHRGLOAD	PORTJbits.RJ3
#define MPULED		PORTHbits.RH0
#define	DIPSW1		PORTDbits.RD0
#define	DIPSW2		PORTDbits.RD1
#define	DIPSW3		PORTDbits.RD2
#define	DIPSW4		PORTDbits.RD3
#define	DIPSW5		PORTDbits.RD4
#define	DIPSW6		PORTDbits.RD5
#define	DIPSW7		PORTDbits.RD6
#define	DIPSW8		PORTDbits.RD7
#define INVERTERON	PORTBbits.RB4
#define	FUSEBLOWN	PORTBbits.RB5
#define	BLINK_RATE	20000
#define JB			TRUE
#define	BATRUN 		3					// 
#define	BATRUNF		1					// faster delay
#define	BATTEST		10					//
#define	BATCHARGE_S	7200				// seconds in  hours
#define BATCAP_S	20					// battery capacity at 20H discharge rate
#define	BATCHARGE_M	43200				// seconds in 12 hours
#define BATCAP_M	200
#define	BATCHARGE_L	43200				// seconds in 12 hours
#define BATCAP_L	200	
#define	TIMEOFF		60					// seconds in 1 min
#define ADC_STEP	488					// mv per dac step
#define	ADC_MULT	344					// voltage div factor
#define	SOLARLOW	4000				// 12000 no cc
#define SOLARHIGH	5000				// 15000 no cc
#define BATTCRIT	11000
#define BATTLOW		12000
#define BATTHIGH	16000
#define BATTFRESH	13200
#define BATTDEAD	4000
#define BATTOLD		3000
#define	BATTRAND	3000
#define BATTBOOST	1000
#define BATTCHECK	300
#define	VENTTIME	1200
#define	LOGGERTIME	30
#define AMP300_OF	521					// dc offset for amploc amp300 sensor raw bit count
#define AMP300_MAX	2380				// Vo at max current
#define AMP300_SEN	8					// mV/A
#define AMP50_OF	514					// dc offset for amploc amp50 sensor raw bit count
#define AMP50_MAX	11400				// Vo at max current
#define AMP50_SEN	23					// mV/A
#define CCCODE1		66
#define CCCODE2		26
#define CCLEN1		10
#define CCLEN2		2
#define ADC0_OFF	1713
#define ADC1_OFF	1630
#define ADC2_OFF	488
#define ADC3_OFF	2423
#define ADC4_OFF	1910
#define ADC5_OFF	488
#define ADC6_OFF	488
#define ADC7_OFF	488
#define	VCP			13200				// Voltage at full charge
#define	CCP			4					// precent of bank cap full charged current.
#define	CTP			4					// time at VCP
#define CCLEDTIME	600					// wait 10 min of solid LED on for charged flag from C40 in FLOAT mode
#define	CCLEDBLINK	5					// seconds to count blink clocks.
#define CCONEBLINK	70
#define CCFOURBLINKS 200				// C40 in absorption mode
#define	CCLEDSOLID	500					// need to check this


volatile int	i=0 ,y=0, z=0, bn=0,brand, FCHECK=TRUE,BATTNUM=4,TIMERFLAG=FALSE,PRIPOWEROK=TRUE,ABSORP=FALSE,CHARGING=FALSE,venttimer=VENTTIME,loggertime=LOGGERTIME,FORCEOUT=FALSE,
				CCDONE=FALSE,CHARGED=FALSE,a50,a300;
volatile long	j=0,alive_led=0,rawint_count=0, timerint_count=0, commint_count=0, status_count=0,cctimer=CCLEDTIME,current,currentin;
volatile 		unsigned long 	vbatol,vbat0,vbat1,vbat2,vbat3,solar,Vin,mV,Vdec,Vfrac,vbatboost=0,rawp1=0,rawp2=0,inputpower=0,primarypower=0;


/*
 * Step #1  The data is allocated into its own section.
 */
#pragma idata bigdata

//				***
//				0.2		solar charge monitor software.
//				0.3		add A/D inputs for voltage/current monitor.
//				0.4		program battery switching and charge monitor
//				0.5		interface and data functions
//				0.6		randomize battery charging
//				0.7		load testing solar
//				0.8		load testing battery
//				0.9		battery data collection 
//				1.0		basic operation 
//				1.1		refine timing and lcd display
//				1.2		DT relays nornally closed power
//				1.3		improve charge logic
//				1.4		display
//				1.5		better battery checks
//				1.6		setup number of batteries
//				1.7		monitor rawp1(cell) rawp2(wind)
//				1.8		monitor charging source(s)
//				1.9		fix battery flags and timers
//				2.0		timer0 isr for flasher and time
//				2.1		rs-232 comms
//				2.2		cleanup for first basic release
//				2.3		format LCD output
//				2.4		status report update for all systems
//				2.5		monitor primary battery #1 (primarypower on ad4)
//				2.6		refactor code
//				2.7		more software cleanup with defines
//				2.8		fix cut/paste error	in critical battery section, add auto data logging
//				2.9		current sensor code for AMPLOC AMP300 sensor
//				3.0		change PORTF usage to PORTB, we now use PORTF as analog inputs.
//				3.1		Change host port to com2 so we can use ttl level inputs from CC on port1
//				3.2		Add FORCEOUT host command and code to check for charge controller status on port1
//				3.3		ADC voltage adjustments from host control
//				3.4		start basic battery monitor fuctions using Victron PMB as a model.
//				3.5		C40 LED status linked to program. two rotines to monitor the LED. one just checks for solid LED, the other counts blinks
//						We will check both as a failsafe mode to make sure the battery is charged. The blink monitor also looks for when we are in
//						ABSORPTION mode and if the C40 in in BULK run mode.
//				3.6		cleanup current monitor and add GPL header.	
//				***
//				dipswitch settings PORTD
//				1	on=reset battery charging counters and flags
//				2	on=2 batteries/off=4 batteries to monitor
//				3	on=allways show countdown for charge time
//				4	on=don't loop waiting for charge power setpoint
//				5	on=stop datalogging 
//				6	on=don't force primary battery charge on low voltage
//				7	on=bypass external charge controller
//				8	on=equalize charge battery

char			lcdstr[18]="SOLARPIC 3.6FB |",bootstr1[18]="Power Up        ",bootstr2[40]="Status: OK     ";
struct	timeruntype		{
				unsigned long	hour,day,weeks;
				};	
struct celltype {
				char			id;											//	battery cell type S,M,L
				unsigned long	voltage,current,charge,cycles,noload,date;
				int				online,discharged,dead,critical,valid,fresh;
				};
struct histtype {
				unsigned int	peak,rate,cef,peukert,soc,ce,ttg,amphour,cb;
				unsigned int hist[10];
				};

struct buttype	{
				unsigned int	bv,bii,bio;
				};
volatile struct	timeruntype		timerun;							
#pragma idata

#pragma udata celldata

volatile struct celltype cell[5];												// data array for each battery cell
	
#pragma udata

#pragma udata	buthdata

volatile struct buttype buth[2][15];											// data array for each battery history data / hour
	
#pragma udata
#pragma udata	butddata

volatile struct buttype butd[2][12];											// data array for each battery history data / day
	
#pragma udata
#pragma udata	butwdata

volatile struct buttype butw[2][14];											// data array for each battery history data / weeks
	
#pragma udata


#pragma	udata battdata

far char		battstatus[220],hms_string[12];
far char		f1[7],f2[7];
struct	battbufftype	{
				unsigned long	length, position;
				int				busy,flag,good;
				};
struct	ccledtype		{
				unsigned char	flag,seconds;
				unsigned int	count,ticks,online;		// >70 ticks = 1 flash, >200 4 to 5 flashes
				};				
			
volatile struct battbufftype	battbuffer={0},batttype={0}, adctype={0};
volatile unsigned char			adc_cal[]={127,127,127,127,127,127,127,127,127};
volatile	struct	ccledtype	ccled_flag={OFF,0,CCLEDBLINK,0,FALSE};
				
#pragma	udata

#pragma	udata histdata

volatile struct histtype	hist[5];

#pragma	udata

#pragma code rx_interrupt = 0x8
void rx_int (void)
{
  _asm goto rx_handler _endasm
}
#pragma code


void	write_data_eeprom(unsigned char data,unsigned char count,unsigned int addr, unsigned int offset)
{
												//  eeprom data array: 0=0x57 checksum, 1=length of array 2=start of array data, array offset
	Busy_eep();
	Write_b_eep(0+offset,0x57);					//	write checksum 0x57 at byte 0 of the offset

	Busy_eep();
	Write_b_eep(1+offset,count);				// length of data

	Busy_eep();
	Write_b_eep(addr+2+offset,data);				//  data
}



#pragma interrupt rx_handler
void rx_handler (void)
{
  unsigned char c1,c2;
  unsigned int dcount,ZAP=FALSE;

	rawint_count++;											// debug counters

	if (PIR1bits.TMR1IF)									// check timer1 irq
	{
		PIR1bits.TMR1IF=0;
		WriteTimer1(0);
		

		if (CCLED==ON) ccled_flag.count++;								// bink speed routine
		if (CCLED==OFF && ccled_flag.flag) {						// led on flag was set but now LED signal is OFF
			ccled_flag.flag=FALSE;
		}


		if (battbuffer.flag && battbuffer.good) {			// check comm flags
			battbuffer.flag=FALSE;
			battbuffer.position=0;
			battbuffer.length=strlen(battstatus);
			battbuffer.busy=TRUE;
		}


		if (battbuffer.busy && battbuffer.good) {			// send data at timer1 irq rate
			if (battbuffer.position<battbuffer.length) {
  				MPULED = !MPULED;						//  flash led
				commint_count++;
				putc2USART(battstatus[battbuffer.position++]);
			} else {
				status_count++;
				battbuffer.busy=FALSE;
			}
		}
	}


   if (INTCONbits.TMR0IF)									// check timer0 irq
   {
 		//check for TMR0 overflow
      	INTCONbits.TMR0IF = 0; 								//clear interrupt flag
      	WriteTimer0(TIMEROFFSET);

      	//Code used by the interrupt goes here
  		MPULED = !MPULED;						//  flash led 
		timerint_count++;									// set 1 second clock counter.
		TIMERFLAG=TRUE;

		if (ccled_flag.seconds-- < 1) {						// count blink ticks in CCLEDBLINK seconds
			ccled_flag.ticks=ccled_flag.count;
			ccled_flag.count=0;
			ccled_flag.seconds=CCLEDBLINK;
			ccled_flag.online=FALSE;
			ABSORP=FALSE;
			if (ccled_flag.ticks>CCONEBLINK) ccled_flag.online=TRUE;
			if (ccled_flag.ticks>CCFOURBLINKS) ABSORP=TRUE;
		}
	
		if (((CCLED==ON) && ccled_flag.flag) && CHARGING && (solar>=VCP) && (ccled_flag.ticks>CCLEDSOLID)) {		// C40 led status routine (LED-lit,charging state, solar at correct voltage and ticks are at the solid level)
																													// overkill but lets really be sure
			if (cctimer-- <1) {
				CHARGED=TRUE;
				CHARGING=FALSE;
				ABSORP=FALSE;
				CHARGERLED=ON;									// turn on relay
			}
		} else {
			cctimer=CCLEDTIME;
			CHARGED=FALSE;
			CHARGERLED=OFF;										// turn off relay
		}
		if (CCLED==ON) ccled_flag.flag=TRUE;											// set led on flag if LED is lit during the 1 second timer check
		
		if ((INVERTERON == 1) || (!PRIPOWEROK) || (CHARGING)) {
			VENTFAN=ON;										// system is active, clear the air
			venttimer=VENTTIME;
		} else {
			if (venttimer-- < 1) VENTFAN=OFF;
		}

		if (STFU)	ALARMOUT=OFF;							// turn off alarm

  		if (DIPSW5==0) {										// check for auto logger to serial port
			if (loggertime-- < 1) {
				battbuffer.flag=TRUE;
				ZAP=TRUE;
				loggertime=LOGGERTIME;
			}
		}
   }


	
  /* Get the character received from the USARTs */

  if (DataRdy1USART()) {					// is data from AUX

		c2 = Read1USART();					// read data from port1
		putc1USART(c2);						// just echo for now

	};



  if (DataRdy2USART()) {					// is data from host

		c1 = Read2USART();					// read from host port2
		if (c1 == 'A') {												// send complete status string.
			battbuffer.flag=TRUE;
			ZAP=TRUE;
		}
		if (c1 == 'F') {												// set charge break flag
			FORCEOUT=TRUE;
		}
		if (c1 == 'K') {												// lockup controller in loop force WDT timeout
			while (TRUE);
		}
		if (c1 == 'B') {												// select battery 1,2,3,4,0=array data
			batttype.position++;
			if (batttype.position > 4) batttype.position=0;
			batttype.busy=TRUE;
		}

		if (c1 == 'D') {												// select ADC to calibrate voltage.
			adctype.position++;
			if (adctype.position > 8) adctype.position=1;
			adctype.busy=TRUE;
			adctype.flag=FALSE;
		}


		if (batttype.busy) {											// battery selected. what operation?
			if (c1 == 'C') {											// does nothing yet
				battbuffer.good=FALSE;
			}															// clear battery eeprom data
			if ((c1 == 'S') || (c1 == 'M') || (c1 == 'L')) {
				cell[batttype.position].id=c1;							// set battery type eeprom data S gell,M agm, L lead-acid
				write_data_eeprom(c1,5,batttype.position,0);
				battbuffer.good=FALSE;
			}
		}

		if (adctype.busy) {											// battery selected. what operation?
			if (c1 == 'V') {
				adctype.flag=TRUE;
			}															// clear battery eeprom data
			if (((c1 == 'M') || (c1 == 'L')) && adctype.flag) {
				if (c1 =='M') {
					adc_cal[adctype.position-1]++;							// set 
					write_data_eeprom(adc_cal[adctype.position-1],8,adctype.position-1,8);
				}
				if (c1 =='L') {
					adc_cal[adctype.position-1]--;							// set 
					write_data_eeprom(adc_cal[adctype.position-1],8,adctype.position-1,8);
				}
			}
		}

		if (c1 == 'Z' || ZAP) {											// clear command flags
			batttype.busy=FALSE;
			batttype.flag=FALSE;
			batttype.good=FALSE;
			batttype.length=0;
			batttype.position=0;
			adctype.busy=FALSE;
			adctype.flag=FALSE;
			adctype.good=FALSE;
			adctype.length=0;
			adctype.position=0;
			ZAP=FALSE;
		}		
  };

	if (((primarypower<BATTLOW) && (primarypower>1000)) && (cell[1].valid == cell[0].valid)) {			// check for low battery condition not disconnected battery


  		if (DIPSW6==0) {								// force charge voltage to primary battery
			if (PRIPOWEROK) {								// went from good to bad
				BCRELAYS=0xFF;									// turn off all relays
				IORELAYS=0xFF;
				putc2USART('?');							// send alert char to host
			}
			BATLOAD=OFF;								// battery load relay/off
			CHRGLOAD=OFF;								// charge load relay/off
			CHRGVOLTS=ON;								// charge voltage relay/on
			BAT1=ON;								// battery #1 relay/off
			CHRG1=ON;								// charge #1 relay/on
			cell[1].date=timerint_count;					// load timer count at start of charging time
		}
		PRIPOWEROK=FALSE;									// set software flag and trigger HOUSE power charger/generator
		CHARGER=ON;									// RB0 is external battery charger/generator relay 	
	} else {
		if ((primarypower>BATTFRESH) && (cell[1].valid == cell[0].valid)) {					// if battery is full recharged
			if (DIPSW6==0) {
				if (!PRIPOWEROK) {							// went from bad to good
					BCRELAYS=0xFF;								// turn off all relays
					IORELAYS=0xFF;
					cell[1].cycles=cell[1].cycles+(timerint_count-cell[1].date);			// update charge time data
					putc2USART('?');						// send alert char to host
				}
			}
			PRIPOWEROK=TRUE;
			CHARGER=OFF;
		}
	}

	if (DIPSW7==1) {									// if set bypass the solar charge controller
		CCBYPASS=ON;									// enable relay
	} else {
		CCBYPASS=OFF;									// turn off relay.
	}

	/* clear interrupt flag */
    PIR1bits.RCIF = 0;
}

unsigned char	read_data_eeprom(unsigned int addr, unsigned int offset)
{
	Busy_eep();
	return Read_b_eep(addr+offset);
}


void wdtdelay(unsigned long int delay)
{
unsigned long int	dcount;

		for (dcount=0; dcount<=delay; dcount++) {	// delay a bit							// all leds on
		ClrWdt();									// reset the WDT timer
	};
}

void lwdtdelay(unsigned long int delay)
{
unsigned long int	dcount;

		for (dcount=0; dcount<=delay; dcount++) {	// delay a bit							// all leds on
		wdtdelay(32000);									// reset the WDT timer
	};
}

void wdttime(unsigned long int delay)
{
unsigned long int	dcount;

  	/* Disable all high priority interrupts */
  	INTCONbits.GIEH = 0;
	TIMERFLAG=FALSE;
  	/* Enable all high priority interrupts */
  	INTCONbits.GIEH = 1;

	while (!TIMERFLAG) {							// sync with timer counter
		ClrWdt();
		dcount=timerint_count;
	}
	
	while (timerint_count<(dcount+delay)) {			// wait until delay
		ClrWdt();
	}
}

void	DelayFor18TCY(void)
{
	Delay10TCYx(72);					// works with 65
}

//------------------------------------------
void DelayPORXLCD (void) {					// works with 15
Delay10KTCYx(15);// Delay of 15ms
		// Cycles = (TimeDelay * Fosc) / 4
		// Cycles = (15ms * 4MHz) / 4
		// Cycles = 15,000
return;
}

//------------------------------------------
void DelayXLCD (void) {						// works with 5
Delay10KTCYx(5); // Delay of 5ms
		// Cycles = (TimeDelay * Fosc) / 4
		// Cycles = (5ms * 4MHz) / 4
		// Cycles = 5,000
return;
}

void putc1(unsigned int c)
{
			 while (Busy1USART()) {};			// wait until the usart is clear
			 putc1USART (c);
}

void putc2(unsigned int c)
{
			 while (Busy2USART()) {};			// wait until the usart is clear
			 putc2USART (c);
}

void start_delay(void)
{
	wdttime(1);									// wait for 1 seconds.
}
	
void MputsXLCD(char *buffer)
{
  while( BusyXLCD() );
  putsXLCD(buffer);
  while( BusyXLCD() );
}

void MputrsXLCD(const rom char *buffer)
{
  while( BusyXLCD() );
  putrsXLCD(buffer);
  while( BusyXLCD() );
}


		
void ADC_read(void)										// update all voltage readings.
{
	int	i;
	OpenADC( ADC_FOSC_64      &
         ADC_RIGHT_JUST   &
         ADC_20_TAD,
         ADC_CH0          &
         ADC_REF_VDD_VSS  &
         ADC_INT_OFF , 12);

	SetChanADC(ADC_CH0);
	Delay10TCYx(250);

	Vin=0;
    for(i=0;i<16;i++)
    {
    	ConvertADC();
    	while(BusyADC());
    	Vin += (unsigned int) ReadADC();
    }
    Vin /= 16;
	CloseADC();

	vbatol=Vin*(ADC0_OFF-127+adc_cal[0]);					//	voltage correction factor
	vbatol=vbatol/100;


	OpenADC( ADC_FOSC_64      &
         ADC_RIGHT_JUST   &
         ADC_20_TAD,
         ADC_CH1          &
         ADC_REF_VDD_VSS  &
         ADC_INT_OFF , 12);

	SetChanADC(ADC_CH1);
	Delay10TCYx(250);

	Vin=0;
    for(i=0;i<16;i++)
    {
    	ConvertADC();
    	while(BusyADC());
    	Vin += (unsigned int) ReadADC();
    }
    Vin /= 16;
	CloseADC();

	solar=Vin*(ADC1_OFF-127+adc_cal[1]);
	solar=solar/100;

	OpenADC( ADC_FOSC_64      &
         ADC_RIGHT_JUST   &
         ADC_20_TAD,
         ADC_CH2          &
         ADC_REF_VDD_VSS  &
         ADC_INT_OFF , 12);

	SetChanADC(ADC_CH5);			// RF0	for input current AMP300
	Delay10TCYx(250);

	Vin=0;
    for(i=0;i<16;i++)
    {
    	ConvertADC();
    	while(BusyADC());
    	Vin += (unsigned int) ReadADC();
    }
    Vin /= 16;
	CloseADC();

	a300=Vin;
	Vin=Vin-AMP300_OF;						// set zero null point
	current=Vin*(ADC5_OFF-127+adc_cal[5]);
	current=current/100;
	current=current/AMP300_SEN;

	OpenADC( ADC_FOSC_64      &
         ADC_RIGHT_JUST   &
         ADC_20_TAD,
         ADC_CH1          &
         ADC_REF_VDD_VSS  &
         ADC_INT_OFF , 12);

	SetChanADC(ADC_CH3);
	Delay10TCYx(250);

	Vin=0;
    for(i=0;i<16;i++)
    {
    	ConvertADC();
    	while(BusyADC());
    	Vin += (unsigned int) ReadADC();
    }
    Vin /= 16;
	CloseADC();

	rawp1=Vin*(ADC3_OFF-127+adc_cal[3]);
	rawp1 /= 100;

	OpenADC( ADC_FOSC_64      &
         ADC_RIGHT_JUST   &
         ADC_20_TAD,
         ADC_CH1          &
         ADC_REF_VDD_VSS  &
         ADC_INT_OFF , 12);

	SetChanADC(ADC_CH4);			// RA4 is digital input RA5 for A/D AN4.
	Delay10TCYx(250);

	Vin=0;
    for(i=0;i<16;i++)
    {
    	ConvertADC();
    	while(BusyADC());
    	Vin += (unsigned int) ReadADC();
    }
    Vin /= 16;
	CloseADC();

	rawp2=Vin*(ADC4_OFF-127+adc_cal[4]);
	rawp2 /= 100;

	inputpower=rawp1;				// set input PV voltage monitor line
	primarypower=rawp2;				// this is battery 1, the primary power storage cell.

	OpenADC( ADC_FOSC_64      &
         ADC_RIGHT_JUST   &
         ADC_20_TAD,
         ADC_CH2          &
         ADC_REF_VDD_VSS  &
         ADC_INT_OFF , 12);

	SetChanADC(ADC_CH2);
	Delay10TCYx(250);

	Vin=0;
    for(i=0;i<16;i++)
    {
    	ConvertADC();
    	while(BusyADC());
    	Vin += (unsigned int) ReadADC();
    }
    Vin /= 16;
	CloseADC();

	a50=Vin;
	Vin=Vin-AMP50_OF;						// set zero null point
	currentin=Vin*(ADC2_OFF-127+adc_cal[2]);
	currentin=currentin/10;								// tenths of an amp	
	currentin=currentin/AMP50_SEN;

}

void	mkbsstring(void)
{
	int		z;
	if (!battbuffer.busy) {

  		/* Disable all high priority interrupts */
  		INTCONbits.GIEH = 0;
		battbuffer.good=FALSE;
  		/* Enable all high priority interrupts */
  		INTCONbits.GIEH = 1;

		ADC_read();
		sprintf(battstatus,"\r\n^^^,1,%u,%u,%u,%u,%u,%u,%lu,%lu,%lu,%lu,%lu,%lu,%i,%i,%li,%li,%lu,%lu,%c,%c,%c,%c,%u",
		PRIPOWEROK,INVERTERON,DIPSW,LATE,LATJ,LATB,timerint_count,commint_count,status_count,rawint_count,vbatol,solar,a50,a300,currentin,current,inputpower,primarypower,
		cell[1].id,cell[2].id,cell[3].id,cell[4].id,ccled_flag.ticks);

		for (z=1;z<5;z++)			// create 
		{
			sprintf(battstatus+strlen(battstatus),",%u,%lu,%lu,%lu",z,cell[z].noload,cell[z].voltage,cell[z].cycles);
		}
		strcatpgm2ram(battstatus,",###");

  		/* Disable all high priority interrupts */
  		INTCONbits.GIEH = 0;
		battbuffer.good=TRUE;
  		/* Enable all high priority interrupts */
  		INTCONbits.GIEH = 1;
	}
}

char* hms(unsigned long sec)							// convert long (seconds) to time string
{
	unsigned int	h,m,s;
	s=sec;
  	h=(s/3600);
  	s=s-(h*3600);
  	m=(s/60);
  	s=s-(m*60);
  	sprintf(hms_string,"%d:%02d:%02d", h, m, s);
	return hms_string;
}

char* voltfp(unsigned long millvolts,char *strprt)		// convert long (millvolts) to voltage string
{
	int				iw=0,ip=0;
	float			voltfrak;
	voltfrak=(float)millvolts/1000;
	iw=(long)((float)voltfrak);
	ip=(long)((float)voltfrak*100)-iw*100;
	sprintf(strprt,"%d.%02d",(int)iw,(int)ip);
	return strprt;	
}


void ChargeRelayOn(int bn)									// switch relays for battery
{
		if (bn == 1) {
		 BAT1=ON;								// battery #1 relay/off
		 CHRG1=ON;								// charge #1 relay/on
		}
		if (bn == 2) {
		 BAT2=ON;								// battery #2 relay/off
		 CHRG2=ON;								// charge #2 relay/on
		}
		if (bn == 3) {
		 BAT3=ON;								// battery #3 relay/off
		 CHRG3=ON;								// charge #3 relay/on
		}
		if (bn == 4) {
		 BAT4=ON;								// battery #4 relay/off
		 CHRG4=ON;								// charge #4 relay/on
		}
		CHARGING=TRUE;
}

void ChargeRelayOff(int bn)
{
		if (bn ==1 ) {
			BAT1=OFF;								// battery #1 relay/on
			CHRG1=OFF;								// charge #1 relay/off
		}
		if (bn ==2 ) {
			BAT2=OFF;								// battery #2 relay/on
			CHRG2=OFF;								// charge #2 relay/off
		}
		if (bn ==3 ) {
			BAT3=OFF;								// battery #3 relay/on
			CHRG3=OFF;								// charge #3 relay/off
		}
		if (bn ==4 ) {
			BAT4=OFF;								// battery #4 relay/on
			CHRG4=OFF;								// charge #4 relay/off
		}
		CHARGING=FALSE;
		ABSORP=FALSE;
}

int ChargeBatt(int bn,int FCHECK, int TIMED)
{
		unsigned long	batcount,BATCHARGE=3600;


		if ((bn) <1 || (bn>BATTNUM)) return 0;

		SOLAROFF=ON;								// toggle off the solar input to reset CC
		CHRGVOLTS=OFF;
		wdttime(BATRUNF);
		CHRGVOLTS=ON;
		SOLAROFF=OFF;

		if (!PRIPOWEROK) {								// skip if we are in primary battery override charging.
			sprintf(bootstr2,"OVERRIDE CHRG #1");
			SetDDRamAddr(0x00);							// move to first line
  			MputsXLCD(bootstr2);
			return 1;
		}
		cell[bn].date=timerint_count;					// load timer count at start of charging time

		ChargeRelayOn(bn);

		sprintf(bootstr2,"Test %i %s      ",bn,hms(cell[bn].cycles+(timerint_count-cell[bn].date)));
		SetDDRamAddr(0x00);				// move to first line
  		MputsXLCD(bootstr2);

/*		battery test routine	*/
		CHRGVOLTS=OFF;									// charge voltage relay/off
		wdttime(BATRUNF);
		ADC_read();										// read unloaded battery voltage from charge line.
		cell[bn].noload=solar;							// set unloaded battery voltage
		BATLOAD=ON;										// battery load relay/on
		wdttime(BATTEST);								// drain the battery
		ADC_read();										// read loaded battery voltage from charge line.
		cell[bn].voltage=solar;							// set loaded battery voltage in cell struct

		cell[bn].critical=FALSE;						// reset battery flags
		cell[bn].discharged=FALSE;
		cell[bn].dead=FALSE;
		cell[bn].fresh=FALSE;

		if (cell[bn].voltage<BATTCRIT) cell[bn].critical=TRUE;	// set battery flags
		if (cell[bn].voltage<BATTLOW) cell[bn].discharged=TRUE;
		if (cell[bn].voltage<BATTDEAD) cell[bn].dead=TRUE;
		if ((cell[bn].noload-cell[bn].voltage)>BATTOLD) cell[bn].dead=TRUE;
		if (cell[bn].voltage>BATTFRESH) cell[bn].fresh=TRUE;

		BATLOAD=OFF;									// battery load relay/off
		CHRGVOLTS=ON;									// charge voltage relay/on
		wdttime(BATRUNF);
/*		end battery test	*/

		brand=rand();

		if (cell[bn].id == 'L') BATCHARGE=BATCHARGE_L;	// set timer per battery type
		if (cell[bn].id == 'M') BATCHARGE=BATCHARGE_M;
		if (cell[bn].id == 'S') BATCHARGE=BATCHARGE_S;

		while (timerint_count<(cell[bn].date+BATCHARGE)) {
			if (!PRIPOWEROK) return 2;					// in override mode exit from normal charge routine
			mkbsstring();								// read voltages from solar cells
			cell[bn].valid=cell[0].valid;
			cell[bn].online=FALSE;
/*			set cell[bn].voltage in battery test routine above	*/
			cell[bn].charge=solar;
			cell[bn].current=currentin;

			if (FCHECK && (!cell[0].critical)) {
				ChargeRelayOff(bn);

				sprintf(bootstr2,"Tested %i %s      ",bn,hms(cell[bn].cycles+(timerint_count-cell[bn].date)));
				SetDDRamAddr(0x00);						// move to first line
  				MputsXLCD(bootstr2);
				voltfp(cell[bn].noload,f1);
				voltfp(cell[bn].voltage,f2);
  				sprintf(bootstr2,"%i,%s,%s V           ",bn,f1,f2);		// loaded voltage drop
  				SetDDRamAddr(0xc0);						// move to second line  
  				MputsXLCD(bootstr2);
				wdttime(BATRUN);
				if (!cell[bn].dead) cell[bn].online=TRUE;				// only good cells should be online
				return 0;								// just checking so return
			}
			if (((cell[bn].voltage >BATTCRIT) && (!cell[0].critical)) && !TIMED) {				// very low voltage will force a full charge by batcount loop
				if (solar>BATTHIGH+vbatboost) {										// and not check for BATTHIGH voltage
					break;
				}
				if (FORCEOUT) {														// Received FOURCEOUT from host, so exit
					FORCEOUT=FALSE;
					break;
				}

				if (CHARGED) {										// charged signal for C40 controller
					CHARGED=FALSE;
					CHARGERLED=OFF;									// turn off relay
					break;
				}

				if (inputpower<SOLARLOW) {							// input is too low to charge
					break;
				}
			}

			if (cell[0].critical) {
				if (inputpower<SOLARLOW) {							// input is too low to charge
					break;
				}
				BCRELAYS=0x0f;										// charge all batteries at once
			}

			if ((DIPSW3==1) || TIMED) {								// switch 3 show countdown for charge time
				sprintf(bootstr2,"Chrg %s left    ",hms((((cell[bn].date+BATCHARGE)-timerint_count))+1));
				SetDDRamAddr(0x00);				// move to first line
  				MputsXLCD(bootstr2);	
			} else {
				voltfp(cell[bn].noload,f1);
				voltfp(cell[bn].voltage,f2);
				sprintf(bootstr2,"C%i-%s %s %s    ",bn,f1,f2,hms(cell[bn].cycles+(timerint_count-cell[bn].date)));
				SetDDRamAddr(0x00);				// move to first line
  				MputsXLCD(bootstr2);
			}
			voltfp(inputpower,f1);
			voltfp(solar,f2);
  			sprintf(bootstr2,"%s,%s,%li A           ",f1,f2,currentin);
  			SetDDRamAddr(0xc0);										// move to second line  
  			MputsXLCD(bootstr2);
			wdttime(BATRUNF);										// reset the WDT timer
			cell[bn].cycles=cell[bn].cycles+(timerint_count-cell[bn].date);	// update charge time data
		};

			voltfp(solar,f2);
			sprintf(bootstr2,"Done %i-%s V          ",bn,f2);
			SetDDRamAddr(0x00);										// move to first line
  			MputsXLCD(bootstr2);
			if (cell[0].critical) {
				cell[0].critical=FALSE;
				BCRELAYS=0xff;
			}

			ChargeRelayOff(bn);

			cell[bn].cycles=cell[bn].cycles+(timerint_count-cell[bn].date);
			if (!cell[bn].dead) cell[bn].online=TRUE;				// only good cells go online
			return 0;
}

void	battalarm(int time, int pulse, int mode)
{
//	alarm code
	ALARMOUT=ON;
}




void main (void)
{
	int				battvolt,ftest=0;
	unsigned int	delayc=0, spinner=0,dc0=256,dc1=256,dc2=256,dc3=256;
	unsigned long	waittime;
	unsigned char	eep_char=0,s0=50,s1=50,s2=50,s3=50,sc0=0,sc1=0,sc2=0,sc3=0,
					RUP=TRUE,GUP=TRUE,BUP=TRUE,AUP=TRUE;


	battbuffer.busy=FALSE;
	battbuffer.flag=FALSE;
	battbuffer.good=TRUE;
	sprintf(battstatus,"\r\n^^^,0,No Status Report,###");


  /* Configure all PORT B,E,H,J pins for output */
  TRISJ = 0;					// status leds
  TRISH = 0;
  TRISE = 0;					//	relay output 1
  TRISB = 0xF0;
  PORTH = 0;
  TRISD = 0xff;					// switch input 1
  TRISA = 0xff;					// analog inputs
  TRISF = 0xff;					// analog inputs
  TRISCbits.TRISC0=1;			// CC led data input
  ADCON1=0x00;  
  
	for (z=1;z<=BATTNUM;z++) {	//	make sure nothings valid until voltage tested.
		cell[z].valid=0;
	}
	cell[0].valid=TIMEROFFSET;

  OpenXLCD( FOUR_BIT & LINES_5X7 );
   
  /* 
   * Open the USART configured as
   * 8N1, 9600 baud, in receive INT mode
   */
  BAUDCON1 |= 0x08;								// 16 bit mode speed register
  BAUDCON2 |= 0x08;								// 16 bit mode speed register
  
  Open1USART (USART_TX_INT_OFF &
             USART_RX_INT_ON &
             USART_ASYNCH_MODE &
             USART_EIGHT_BIT &
             USART_CONT_RX &
             USART_BRGH_LOW, 8332);				// 40mhz osc HS		300 baud, 16bit divider
												
	SPBRGH1=0x20;
	SPBRG1=0x8C;
             
  Open2USART (USART_TX_INT_OFF &
             USART_RX_INT_ON &
             USART_ASYNCH_MODE &
             USART_EIGHT_BIT &
             USART_CONT_RX &
             USART_BRGH_LOW, 259);				// 40mhz osc HS		9600 baud, 16bit divider
												
	SPBRGH2=0x01;
	SPBRG2=0x03;


    OpenTimer0(TIMER_INT_ON & T0_16BIT & T0_SOURCE_INT & T0_PS_1_256);
    OpenTimer1(TIMER_INT_ON & T1_16BIT_RW & T1_SOURCE_INT & T1_PS_1_1 & T1_OSC1EN_OFF & T1_SYNC_EXT_OFF );

  	WriteTimer0(TIMEROFFSET);						//	start timer0 at 1 second ticks
  	WriteTimer1(0);									// 	comm timer high speed ticks


  /* Enable interrupt priority */
  RCONbits.IPEN = 1;

  /* Make receive interrupt high priority */
  IPR1bits.RCIP = 1;

  /* Enable all high priority interrupts */
  INTCONbits.GIEH = 1;
		
 while (DataRdy1USART()) {		// dump 1 rx data`
	z = Read1USART();
 };
 while (DataRdy2USART()) {		// dump 2 rx data
	z = Read2USART();
 };




  IORELAYS=0xff;								// set leds to off at powerup/reset
  CHRGVOLTS=ON;									// charge voltage isolation relay/on
  BATLOAD=OFF;									// battery load test relay/off
  BCRELAYS=0xFF;								// turn off relay 0-3	battery power
  EXTIO=0x0F;									// turn of external/alarm signals.
  
  SetDDRamAddr(0x00);							// move to first line
  MputsXLCD(bootstr1);  
  SetDDRamAddr(0xc0);							// move to second line
  MputsXLCD(bootstr2);
  start_delay();								// wait for * to powerup and be ready
  SetDDRamAddr(0x00);							// move to first line  
  MputsXLCD(lcdstr);

  if (DIPSW2==1) {								// set number of batteries to 2
	BATTNUM=2;
  }
	  											// leds from outputs to ground via resistor.

	for (z=0;z<=BATTNUM;z++) {
  		if (DIPSW1==1) {						// clear battery data
  	 		cell[z].id='L';
  			INTCONbits.GIEH = 0;
			write_data_eeprom('L',5,z,0);
  			INTCONbits.GIEH = 1;
  		}
		cell[z].voltage=0;
		cell[z].current=0;
		cell[z].charge=0;
  	 	cell[z].cycles=0;
		cell[z].noload=0;
		cell[z].date=0;
  	 	hist[z].peak=0;
		hist[z].rate=0;
		cell[z].online=FALSE;
		cell[z].discharged=FALSE;
		cell[z].dead=FALSE;
		cell[z].critical=FALSE;
		cell[z].fresh=FALSE;
	}

	if (DIPSW1==1) {						// clear ADC cal data
		for (z=0;z<=8;z++) {
      		INTCONbits.GIEH = 0;
			write_data_eeprom(127,8,z,8);
  			INTCONbits.GIEH = 1;
		}
	}

	/*
	 * load EEPROM data
	*/
	if (read_data_eeprom(0,0) == 0x57) {	// load battery types from EEPROM
		for (z=0;z<=BATTNUM;z++) {
			eep_char=read_data_eeprom(2,z);
			if ((eep_char == 'S')||(eep_char == 'M')||(eep_char == 'L')) {
				cell[z].id=eep_char;
			} else {
				cell[z].id='L';
			}             
		}
	}

	if (read_data_eeprom(0,8) == 0x57) {	// load ADC cal data from EEPROM
		for (z=0;z<8;z++) {
			adc_cal[z]=read_data_eeprom(10,z);     
		}
	}

  if (DIPSW8==1) vbatboost=BATTBOOST;						// equalize charge battery
	CHRGVOLTS=ON;											// charge voltage relay/on
  	CHRGLOAD=ON;											// charge load relay/on
  	wdttime(BATRUNF);
	ADC_read();

	voltfp(inputpower,f1);
	voltfp(solar,f2);
	sprintf(bootstr2,"%i b,%s,%s SW   ",BATTNUM,f1,f2);	// check stuff
	SetDDRamAddr(0x00);				// move to first line
  	MputsXLCD(bootstr2);
  	sprintf(bootstr2,"%i,%i,%i,%i,%i,%i,%i,%i           ",DIPSW1,DIPSW2,DIPSW3,DIPSW4,DIPSW5,DIPSW6,DIPSW7,DIPSW8);
  	SetDDRamAddr(0xc0);				// move to second line  
  	MputsXLCD(bootstr2);
	start_delay();

  BCRELAYS=0xFF;										// turn on all battery relays/on
  CHRGLOAD=ON;											// charge load relay/on

  wdttime(BATRUNF);										// read battery and charging system voltages
  ADC_read();
  srand((unsigned int)vbatol);							// set random seed
  cell[0].valid=rand();									// set valid value


  /* Loop forever */  
  while (TRUE)
 {
	if (PRIPOWEROK) {
		BCRELAYS=0xFF;									// turn on all battery relays/on
		CHRGLOAD=ON;									// charge load relay/on

		wdttime(BATRUNF);								// read battery and charging system voltages
		mkbsstring();

		bn=0;											// set battery number 0=battery array
		cell[bn].online=TRUE;
		cell[bn].voltage=vbatol;
		cell[bn].noload=vbatol;
		cell[bn].charge=solar;
		cell[bn].current=currentin;
		cell[bn].cycles++;
		if (cell[bn].voltage<BATTCRIT) cell[bn].critical=TRUE;
		if (cell[bn].voltage<BATTLOW) FCHECK=TRUE;			// check array for low voltage
		for (z=0;z<=BATTNUM;z++) {
			if (cell[z].valid==cell[0].valid) {
				voltfp(cell[z].charge,f1);
				sprintf(bootstr2,"%i-B,V,C %s %s      ",z,f1,hms(cell[z].cycles));
				SetDDRamAddr(0x00);				// move to first line
  				MputsXLCD(bootstr2);
				voltfp(cell[z].noload,f1);
				voltfp(cell[z].voltage,f2);
  				sprintf(bootstr2,"%s,%s,%li A          ",f1,f2,cell[z].current);
  				SetDDRamAddr(0xc0);				// move to second line  
  				MputsXLCD(bootstr2);
				wdttime(BATRUN);
			}
		}

		ftest++;
	}
	if (ftest >BATTCHECK) {							// force battery test after X loops
		FCHECK=TRUE;
		ftest=0;
	}

	if (!PRIPOWEROK) {								// skip if we are in primary battery override charging.
		sprintf(bootstr2,"OVERRIDE CHRG #1");
		SetDDRamAddr(0x00);							// move to first line
  		MputsXLCD(bootstr2);
	}
  	sprintf(bootstr2,"                        ");	//	blank second line
  	SetDDRamAddr(0xc0);								// move to second line  
  	MputsXLCD(bootstr2);

	while ((inputpower > SOLARLOW) || FCHECK)
	{

  	brand=rand();
	if (FCHECK) {
		for (z=1;z<=BATTNUM;z++) {
			if (PRIPOWEROK) {
				CHRGLOAD=ON;									// charge load relay/on	
				wdttime(BATRUNF);
				mkbsstring();
				CHRGLOAD=OFF;									// charge load relay/off
				bn=z;											// set battery #
				ChargeBatt(bn,TRUE,FALSE);						//	check battery 
			}
		}
	}
	FCHECK=FALSE;

	if (PRIPOWEROK) {
		sprintf(bootstr2,"Any critical      ");
		SetDDRamAddr(0x00);				// move to first line
  		MputsXLCD(bootstr2);
  		sprintf(bootstr2,"Batteries?              ");
  		SetDDRamAddr(0xc0);				// move to second line  
  		MputsXLCD(bootstr2);

	for (z=1;z<=BATTNUM;z++) {
			CHRGLOAD=ON;									// charge load relay/on	
  			sprintf(bootstr2,"Battery #%i          ",z);
  			SetDDRamAddr(0xc0);								// move to second line  
  			MputsXLCD(bootstr2);
			wdttime(BATRUNF);
			mkbsstring();

			if (cell[z].critical) {
				if ((inputpower>SOLARHIGH)) {
					CHRGLOAD=OFF;							// charge load relay/off
					bn=z;									// set battery #
					ChargeBatt(bn,FALSE,TRUE);
					ChargeBatt(bn,TRUE,TRUE);
					ftest=BATTCHECK;
				}
			}
		}
	}

	if (PRIPOWEROK) {
		sprintf(bootstr2,"Any low      ");
		SetDDRamAddr(0x00);				// move to first line
  		MputsXLCD(bootstr2);
  		sprintf(bootstr2,"Batteries?                ");
  		SetDDRamAddr(0xc0);				// move to second line  
  		MputsXLCD(bootstr2);

	for (z=1;z<=BATTNUM;z++) {
			CHRGLOAD=ON;									// charge load relay/on
  			sprintf(bootstr2,"Battery #%i          ",z);
  			SetDDRamAddr(0xc0);				// move to second line  
  			MputsXLCD(bootstr2);	
			wdttime(BATRUNF);
			mkbsstring();
			
			if (cell[z].discharged) {
				if ((inputpower>SOLARHIGH)) {
					CHRGLOAD=OFF;									// charge load relay/off
					bn=z;											// set battery #
					ChargeBatt(bn,FALSE,TRUE);
					ChargeBatt(bn,TRUE,TRUE);
					ftest=BATTCHECK;
				}
			}
		}
	}

	if (PRIPOWEROK) {
		sprintf(bootstr2,"Any discharged        ");
		SetDDRamAddr(0x00);				// move to first line
  		MputsXLCD(bootstr2);
  		sprintf(bootstr2,"Batteries?           ");
  		SetDDRamAddr(0xc0);				// move to second line  
  		MputsXLCD(bootstr2);

	for (z=1;z<=BATTNUM;z++) {
			CHRGLOAD=ON;									// charge load relay/on	
  			sprintf(bootstr2,"Battery #%i          ",z);
  			SetDDRamAddr(0xc0);								// move to second line  
  			MputsXLCD(bootstr2);

			wdttime(BATRUNF);
			mkbsstring();

			if (!cell[z].fresh) {
				if ((inputpower>SOLARHIGH)) {
					CHRGLOAD=OFF;									// charge load relay/off
					bn=z;											// set battery #
					ChargeBatt(bn,FALSE,FALSE);
					ChargeBatt(bn,TRUE,FALSE);
					ftest=BATTCHECK;
				}
			}
		}
	}

	ftest++;
	if (ftest >=BATTCHECK) {						// force battery test after X loops
		FCHECK=TRUE;
		ftest=0;
	}


	wdttime(BATRUN);
	for (z=0;z<=BATTNUM;z++) {
		if (cell[z].valid==cell[0].valid) {
			if (!PRIPOWEROK) {
				sprintf(bootstr2,"OVERRIDE CHRG #1");
				SetDDRamAddr(0x00);								// move to first line
  				MputsXLCD(bootstr2);
				mkbsstring();									// update primary battery voltage
				cell[1].voltage=primarypower;					// update cell[1] voltages for display and status
				cell[1].noload=primarypower;
				wdttime(BATRUN);
			}
			voltfp(cell[z].charge,f1);			
			sprintf(bootstr2,"%i-B,V,C %s %s      ",z,f1,hms(cell[z].cycles));
			SetDDRamAddr(0x00);				// move to first line
  			MputsXLCD(bootstr2);
			voltfp(cell[z].noload,f1);
			voltfp(cell[z].voltage,f2);
  			sprintf(bootstr2,"%s,%s,%li A          ",f1,f2,cell[z].current);
  			SetDDRamAddr(0xc0);				// move to second line  
  			MputsXLCD(bootstr2);
			if (cell[z].dead) battalarm(0,0,0);			// possible dead battery
			wdttime(BATRUN);
		}
	}

	}

	waittime=timerint_count;
	while (inputpower<SOLARLOW && (DIPSW4==0)) {			// wait until we have charging power or dipswitch 4 on
			sprintf(bootstr2,"Charge Power Low        ");
			SetDDRamAddr(0x00);				// move to first line
  			MputsXLCD(bootstr2);
			voltfp(inputpower,f1);
			voltfp(solar,f2);
  			sprintf(bootstr2,"%s,%s %s               ",f1,f2,hms(((timerint_count-waittime))));
  			SetDDRamAddr(0xc0);				// move to second line  
  			MputsXLCD(bootstr2);
			wdttime(BATRUN);
			mkbsstring();						// update
			FCHECK=TRUE;
			ftest=0;
	}

	ClrWdt();								// reset the WDT timer
}
}